#! /usr/bin/python2.7
# -*- coding: UTF-8 -*-

#############################################################################
#
# Copyright (C) Labomedia February 2012
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
#############################################################################

# More at
# http://wiki.labomedia.org/index.php/Touch_Osc#Utiliser_TouchOSC_dans_un_script_python_pour_faire_de_la_musique_Kollectiv

# Help: convert msg in dict from
# http://wiki.labomedia.org/index.php/OSCinBlender:_une_aide_pratique

import time
import threading
import socket
import sys
sys.path.append("./OSC")
import OSC
sys.path.append("./pyFluidSynth")

# Set global value
# Dictionary to store all previous received value
previous = {}
# Dictionary to store all variables and values
msgDict = {}
# Bool to stop listen loop
stop_listen = False
# Bool to stop keyboard loop
stop_keyboard = False
# Dictionary to store who is the player with IP
who_dict = {}


def main():
    # Enter ip and port
    test = False # test true to debug quickly
    global ip, unique_port, buff_size
    if not test:
        print 'Sorry, no control input !\n'
        ip = raw_input('What is your IP ? ')
        unique_port = int(raw_input('Unique port number ? '))
    else:
        ip = '10.0.0.4'
        unique_port = 8000

    buff_size = 1024

    print 'Your IP is', ip , 'and Port is', unique_port
    # Set and plug
    plug()
    set_audio()
    print 'No problem with --> fluidsynth: warning:\n'
    print 'Press q + Enter to quit\n'

    # Thread listen
    global thread1
    thread1=threading.Thread(target=thread_listen)
    thread1.start()

    # Thread keyboard quit with q
    global thread2
    thread2=threading.Thread(target=thread_keyboard)
    thread2.start()

def thread_listen():
    '''Always listen UDP until break with stop_listen'''
    while not stop_listen:
        listen()
        if stop_keyboard:
            print 'Exit thread listen'
            # System exit
            exit(0)
            
def thread_keyboard():
    '''Wait on q on keyboard until break with stop_keyboard'''
    global stop_keyboard
    while not stop_keyboard:
        c = sys.stdin.read(1)
        if c == 'q':  # exit program
            print 'Exit thread keyboard'
            # Stop thread_listen
            stop_listen = True
            # Stop this loop
            stop_keyboard = True
            # Join thread1 = listen and stop
            thread1.join()

def plug():
    '''Plug UDP'''
    global socketOSC
    try:
        # Set UDP socket
        socketOSC = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        socketOSC.bind((ip, unique_port))
        socketOSC.setblocking(0)
        socketOSC.settimeout(0.002)
        print 'Plug : IP =', ip, 'Port =', unique_port
    except:
        print 'No connected to IP =', ip, 'Port =', unique_port
        pass

def listen():
    '''Listen UDP and decode OSC'''
    try:
        time.sleep(0.001)
        raw_data, who_are_you = socketOSC.recvfrom(buff_size)
        data = OSC.decodeOSC(raw_data)
        if who_are_you in who_dict:
            who = who_dict[who_are_you]
        else:
            who_dict[who_are_you] = len(who_dict)
            who = who_dict[who_are_you]
        convert_data(data, who)
        use_data()
    except socket.error:
        pass

def convert_data(data, n):
    '''Convert OSC message in variable'''
    # if list
    if isinstance(data, list):
        if len(data) > 1 :
            # if bundle cut bundle and time
            if data[0]=='#bundle':
                # Cut the two first item
                data = data[2:]
            test_if_list_of_list(data, n)
    else:
        pass

def test_if_list_of_list(data, n):
    '''test if first in list is a list'''
    if isinstance(data[0], list):
        for i in range(len(data)):
            list_to_var(data[i])
    else:
        list_to_var(data, n)

def name(s, n):
    '''Convert s in variable name'''
    # Cut the first / if
    if s[0] == '/':
        s = s[1:]
    # add player number in front to identify the player
    # so the first item is never a number
    s = 'n' + str(n) + '_' + s
    # Replace / or - or , with _ (add str in list if necessary)
    for i in ['/', '-', ',']:
        s = s.replace(i, '_')
    return s
    
def list_to_var(data, n):
    '''Set value in variable'''
    # First item is the tag, otherwise message in not osc
    var_name = name(data[0], n)
    # if the second is type of all values, example 'iif
    if isinstance(data[1], str) or data[1][0] == ",":
        # cut the 2 first
        data = data[2:]
    # cut the first: msg don't respect OSC format
    else:
        data = data[1:]
    # data is always list, convert list to int or float or str if len=1
    if len(data)==1:
        data = data[0]
    msgDict[var_name] = data

def use_data():
    '''Play if new data'''
    # go to what else if new value in msgDict
    for event, value in msgDict.items():
        if event in previous:
            if previous[event] != value:
                previous[event] = value
                if event.find('_z') == -1:
					spam(event)
        else:
            previous[event] = value
            if event.find('_z') == -1:
				spam(event)

def spam(event):
    '''Who play and what '''
    # example: n1_ where 1 is the player
    player = event.split('_')[0] + '_'

    # channel
    channel = 1
    for j in [1, 3]:
        if player + str(j) + '_fader1' in msgDict:
            channel = 1 + int(15*msgDict[player + str(j) + '_fader1'])

    # volume
    volume = 100
    for j in [1, 3]:
        if player + str(j) + '_fader2' in msgDict:
            volume = 47 + int(80*msgDict[player + str(j) + '_fader2'])

    # life
    life = 0.4
    if player+'3_rotary1' in msgDict:
        life = 2*msgDict[player+'3_rotary1']

    # note with 1_push
    if event.find(player+'1_push') != -1:
        if int(msgDict[event]) == 1:
            note = 47+int(event[9:])
            # play if event is note
            thread_note(note, life, channel, volume)

    # note with 2_push
    if event.find(player+'2_push') != -1:
        if int(msgDict[event]) == 1:
            note = 59+int(event[9:])
            # play if event is note
            thread_note(note, life, channel, volume)

    # note with multitoggle
    if event.find(player+'2_multitoggle') != -1:
        if int(msgDict[event]) == 1:
            note = 12+int(event.split("_")[4])+((int(event.split("_")[3])-1)*16)
            # play if event is note
            thread_note(note, life, channel, volume)

    # note with 4_xy
    if event.find(player+'4_xy') != -1:
        note = 24+int(100*msgDict[event][1])
        volume = 50 + int(50*msgDict[event][0])
        # play if event is note
        thread_note(note, life, channel, volume)
        
def set_audio():
    '''Set fluidsynth, be carefull with fonts path'''
    try:
        import fluidsynth
        global fs
        fs = fluidsynth.Synth()
        fs.start(driver='alsa')
        # This path is a problem because it's absolute
        sfid = fs.sfload("/usr/share/sounds/sf2/FluidR3_GM.sf2")

        # .program_select(channel,    sfid,   bank,   bank number)
        fs.program_select(  1  ,      sfid,   0,      0  ) # yamaha grand piano
        fs.program_select(  2  ,      sfid,   0,      40 ) # violin
        fs.program_select(  3  ,      sfid,   0,      37 ) # pop bass
        fs.program_select(  4  ,      sfid,   0,      56 ) # trompet
        fs.program_select(  5  ,      sfid,   0,      66 ) # tenor sax
        fs.program_select(  6  ,      sfid,   0,      114) # steel drums
        fs.program_select(  7  ,      sfid,   0,      118) # synth drum
        fs.program_select(  8  ,      sfid,   0,      119) # reverse cymbal
        fs.program_select(  9  ,      sfid,   0,      116) # taiko drum
        fs.program_select(  10  ,     sfid,   0,      73 ) # flute
        fs.program_select(  11  ,     sfid,   0,      70 ) # basson
        fs.program_select(  12  ,     sfid,   0,      46 ) # harp
        fs.program_select(  13  ,     sfid,   0,      13 ) # xylophone
        fs.program_select(  14  ,     sfid,   0,      24 ) # nylon string guitar
        fs.program_select(  15  ,     sfid,   0,      25 ) # steel string guitar
        fs.program_select(  16  ,     sfid,   0,      29 ) # overdrive guitar

    except:
        print "For sound synthetizer, you must install:"
        print "sudo apt-get install fluidsynth fluid-soundfont-gm fluid-soundfont-gs"
        print "Be carefull with Sound Font Path"
        pass

def playNote(note, life, channel, volume):
    '''note from 0 to 127 but all values are not possible in all bank
    life 0 to ?
    channel 1 to 16
    volume 0 to 127'''
    fs.noteon(channel, note, volume)
    time.sleep(life)
    fs.noteoff(channel, note)

def thread_note(note, life, channel, volume):
    thread3 = threading.Thread(target=playNote, args=(note, life, channel, volume))
    thread3.start()
    #print "Play note =", note, "Volume =", volume, "Channel =", channel, "Life =", life

# Run main
main()
